﻿using System;
using System.Collections.Generic;
using System.Drawing;
using SharpDX.Direct3D9;
using SharpDX.XInput;
using SharpDX.RawInput;
using App = RVII.RVIIManager;

namespace RVII
{
    public static class GSConv
    {
        public static SharpDX.ColorBGRA Color(Color c)
        {
            return new SharpDX.ColorBGRA(c.R, c.G, c.B, c.A);
        }

        public static SharpDX.Rectangle Rectangle(Rectangle r)
        {
            return new SharpDX.Rectangle(r.Left, r.Top, r.Right, r.Bottom);
        }

        public static SharpDX.Vector2 Vector2(Point p)
        {
            return new SharpDX.Vector2(p.X, p.Y);
        }

        public static SharpDX.Vector3 Vector3(Point p)
        {
            return new SharpDX.Vector3(p.X, p.Y, 0.0f);
        }
    }

    public enum StringAlignment
    {
        Near, Center, Far
    }

    public static class Visual
    {
        public static void Paint(GraphicsObjects go, Rectangle bounds, Color c)
        {
            go.SpriteBatch.Transform = SharpDX.Matrix.Scaling(bounds.Width, bounds.Height, 1.0f) * SharpDX.Matrix.Translation(GSConv.Vector3(bounds.Location));
            go.SpriteBatch.Draw(App.Global.PlaneTex, GSConv.Color(c));
            go.SpriteBatch.Transform = SharpDX.Matrix.Identity;
        }

        private class SpriteTextData
        {
            public Texture Tex { get; set; }
            public bool Used { get; set; }
        }

        private struct SpriteTextKey : IEquatable<SpriteTextKey>
        {
            public string Text;
            public System.Drawing.Font Font;

            public SpriteTextKey(string t, System.Drawing.Font f)
            {
                Text = t;
                Font = f;
            }

            public bool Equals(SpriteTextKey other)
            {
                return Text.Equals(other.Text) && Font.Equals(other.Font);
            }
        }

        private static Surface parent_rt = null;
        private static Surface wnd_rt = null;
        private static Dictionary<SpriteTextKey, SpriteTextData> tmap = new Dictionary<SpriteTextKey, SpriteTextData>(128);

        internal static void TrimFontTexCache()
        {
            bool done;
            do {
                done = false;
                foreach (KeyValuePair<SpriteTextKey, SpriteTextData> kv in tmap) {
                    if (!kv.Value.Used) {
                        kv.Value.Tex.Dispose();
                        tmap.Remove(kv.Key);
                        done = true;
                        break;
                    }
                }
            } while (done);
            foreach (KeyValuePair<SpriteTextKey, SpriteTextData> kv in tmap) {
                kv.Value.Used = false;
            }
        }

        internal static void ClearFontTexCache()
        {
            if (tmap != null) {
                foreach (KeyValuePair<SpriteTextKey, SpriteTextData> kv in tmap) {
                    if (kv.Value != null && kv.Value.Tex != null) {
                        kv.Value.Tex.Dispose();
                    }
                }
                tmap.Clear();
            }
        }

        private static int NormalizeLengthForTexture(int length)
        {
            int result = 1;
            while (length > result) { result *= 2; }
            return result;
        }

        private static void DrawString(GraphicsObjects go, System.Drawing.Font f, string text, SharpDX.Vector2 location, Color textColor)
        {
            if (text.Length == 0) return;

            SpriteTextKey tk = new SpriteTextKey(text, f);
            SpriteTextData td;
            if (!tmap.TryGetValue(tk, out td)) {
                Size s = System.Windows.Forms.TextRenderer.MeasureText(text, f);
                s.Width = NormalizeLengthForTexture(s.Width);
                s.Height = NormalizeLengthForTexture(s.Height);
                byte[] c = new byte[s.Width * s.Height * 4];
                using (Bitmap b = new Bitmap(s.Width, s.Height)) {
                    using (Graphics g = Graphics.FromImage(b)) {
                        g.Clear(Color.Empty);
                        System.Windows.Forms.TextRenderer.DrawText(g, text, f, Point.Empty, Color.White);
                    }
                    System.Drawing.Imaging.BitmapData d = b.LockBits(
                        new Rectangle(0, 0, s.Width, s.Height),
                        System.Drawing.Imaging.ImageLockMode.ReadOnly,
                        System.Drawing.Imaging.PixelFormat.Format32bppArgb);
                    try {
                        System.Runtime.InteropServices.Marshal.Copy(d.Scan0, c, 0, c.Length);
                    } finally {
                        b.UnlockBits(d);
                    }
                }
                td = new SpriteTextData();
                td.Tex = new Texture(go.SpriteBatch.Device, s.Width, s.Height, 1, Usage.None, Format.A8R8G8B8, Pool.Managed);
                SharpDX.DataRectangle dr = td.Tex.LockRectangle(0, LockFlags.None);
                for (int i = 0; i < s.Height; i++) {
                    System.Runtime.InteropServices.Marshal.Copy(c, 4 * s.Width * i, (IntPtr)(dr.DataPointer.ToInt64() + dr.Pitch * i), dr.Pitch);
                }
                td.Tex.UnlockRectangle(0);
                tmap.Add(tk, td);
            }
            go.SpriteBatch.Draw(td.Tex, GSConv.Color(textColor), null, null, new SharpDX.Vector3(location, 0.0f));
            td.Used = true;
        }

        public static void DrawString(GraphicsObjects go, FontInfo f, string text, Point location)
        {
            DrawString(go, f.Instance, text, GSConv.Vector2(location), Color.White);
        }

        public static void DrawString(GraphicsObjects go, FontInfo f, string text, Point location, Color c)
        {
            DrawString(go, f.Instance, text, GSConv.Vector2(location), c);
        }

        public static void DrawString(GraphicsObjects go, FontInfo f, string text, Rectangle bounds, Color c, StringAlignment vertical, StringAlignment horizontal)
        {
            Size s = System.Windows.Forms.TextRenderer.MeasureText(text, f.Instance);

            SharpDX.Vector2 wh = new SharpDX.Vector2(s.Width, s.Height); //go.SmallFont.MeasureString(text);
            switch (vertical) {
                case StringAlignment.Near:
                    wh.Y = bounds.Y;
                    break;
                case StringAlignment.Center:
                    wh.Y = bounds.Y + (bounds.Height - wh.Y) / 2;
                    break;
                case StringAlignment.Far:
                    wh.Y = bounds.Y + bounds.Height - wh.Y;
                    break;
            }
            switch (horizontal) {
                case StringAlignment.Near:
                    wh.X = bounds.X;
                    break;
                case StringAlignment.Center:
                    wh.X = bounds.X + (bounds.Width - wh.X) / 2;
                    break;
                case StringAlignment.Far:
                    wh.X = bounds.X + bounds.Width - wh.X;
                    break;
            }

            DrawString(go, f.Instance, text, wh, c);
        }

        public static void BeginWindowMode(GraphicsObjects go)
        {
            /*Should call with try-finally block.
             * BeginWindowMode(go);
             * try {
             *   DrawToWindow();
             * } finally {
             *   EndWindowMode(go, bounds, c);
             * }
             */
            if (parent_rt == null) {
                go.SpriteBatch.End();
                parent_rt = go.SpriteBatch.Device.GetRenderTarget(0);
                wnd_rt = App.Global.WndTex.GetSurfaceLevel(0);
                go.SpriteBatch.Device.SetRenderTarget(0, wnd_rt);
                go.SpriteBatch.Device.Clear(ClearFlags.Target, GSConv.Color(Color.Empty), 1.0f, 0);
                go.SpriteBatch.Begin(SpriteFlags.AlphaBlend | SpriteFlags.DoNotSaveState);
            } else {
                throw new InvalidOperationException("Visual.BeginWindowMode is already called.");
            }
        }

        public static void EndWindowMode(GraphicsObjects go, Rectangle bounds, Color c)
        {
            if (parent_rt != null) {
                go.SpriteBatch.End();
                go.SpriteBatch.Device.SetRenderTarget(0, parent_rt);
                go.SpriteBatch.Begin(SpriteFlags.AlphaBlend | SpriteFlags.DoNotSaveState);
                go.SpriteBatch.Draw(App.Global.WndTex, GSConv.Color(c), GSConv.Rectangle(new Rectangle(0, 0, bounds.Width, bounds.Height)), null, GSConv.Vector3(bounds.Location));
                wnd_rt.Dispose();
                wnd_rt = null;
                parent_rt.Dispose();
                parent_rt = null;
            } else {
                throw new InvalidOperationException("Visual.BeginWindowMode is not called yet.");
            }
        }
    }

    public static class InputUtil
    {
        private struct ButtonSetting
        {
            public byte Map;
            public byte Interval1;
            public byte Interval2;
            public byte IntervalRef;
        }

        private static ButtonSetting[] button_settings = new ButtonSetting[32];
        private static byte[] key_button_map = new byte[64];
        private static int[] state_buffer = new int[60];
        private static bool[] global_key_state = new bool[256];

        private static bool IsButtonDown(Gamepad gps, int button)
        {
            switch (button) {
                case 1:
                    return (gps.Buttons & GamepadButtonFlags.A) != 0;
                case 2:
                    return (gps.Buttons & GamepadButtonFlags.B) != 0;
                case 3:
                    return (gps.Buttons & GamepadButtonFlags.X) != 0;
                case 4:
                    return (gps.Buttons & GamepadButtonFlags.Y) != 0;
                case 5:
                    return (gps.Buttons & GamepadButtonFlags.LeftShoulder) != 0;
                case 6:
                    return (gps.Buttons & GamepadButtonFlags.RightShoulder) != 0;
                case 7:
                    return (gps.Buttons & GamepadButtonFlags.Back) != 0;
                case 8:
                    return (gps.Buttons & GamepadButtonFlags.Start) != 0;
                case 9:
                    return (gps.Buttons & GamepadButtonFlags.LeftThumb) != 0;
                case 10:
                    return (gps.Buttons & GamepadButtonFlags.RightThumb) != 0;
                case 32:
                    return false; // gps.IsButtonDown(Buttons.BigButton);
                case 33:
                    return (gps.Buttons & GamepadButtonFlags.DPadUp) != 0;
                case 34:
                    return (gps.Buttons & GamepadButtonFlags.DPadRight) != 0 &&
                        (gps.Buttons & GamepadButtonFlags.DPadUp) != 0;
                case 35:
                    return (gps.Buttons & GamepadButtonFlags.DPadRight) != 0;
                case 36:
                    return (gps.Buttons & GamepadButtonFlags.DPadDown) != 0 &&
                        (gps.Buttons & GamepadButtonFlags.DPadRight) != 0;
                case 37:
                    return (gps.Buttons & GamepadButtonFlags.DPadDown) != 0;
                case 38:
                    return (gps.Buttons & GamepadButtonFlags.DPadLeft) != 0 &&
                        (gps.Buttons & GamepadButtonFlags.DPadDown) != 0;
                case 39:
                    return (gps.Buttons & GamepadButtonFlags.DPadLeft) != 0;
                case 40:
                    return (gps.Buttons & GamepadButtonFlags.DPadUp) != 0 &&
                        (gps.Buttons & GamepadButtonFlags.DPadLeft) != 0;
                case 41:
                    return gps.LeftThumbX <= -0x4000;
                case 42:
                    return gps.LeftThumbX >= 0x4000;
                case 43:
                    return gps.LeftThumbY >= 0x4000;
                case 44:
                    return gps.LeftThumbY <= -0x4000;
                case 45:
                    return gps.LeftTrigger >= 0x40;
                case 46:
                    return gps.RightTrigger >= 0x40;
                case 49:
                    return gps.RightThumbX <= -0x4000;
                case 50:
                    return gps.RightThumbX >= 0x4000;
                case 51:
                    return gps.RightThumbY >= 0x4000;
                case 52:
                    return gps.RightThumbY <= -0x4000;
            }
            return false;
        }

        private static void KeyboardInputProc(object sender, KeyboardInputEventArgs args)
        {
            if ((args.ScanCodeFlags & ScanCodeFlags.Break) == 0) {
                global_key_state[(int)args.Key] = true;
            } else {
                global_key_state[(int)args.Key] = false;
            }
        }

        internal static void RegisterKeyboard()
        {
            SharpDX.RawInput.Device.RegisterDevice(SharpDX.Multimedia.UsagePage.Generic, SharpDX.Multimedia.UsageId.GenericKeyboard, DeviceFlags.NoLegacy, App.Global.Window.Handle, true);
            SharpDX.RawInput.Device.KeyboardInput += KeyboardInputProc;
        }

        internal static void UnregisterKeyboard()
        {
            SharpDX.RawInput.Device.KeyboardInput -= KeyboardInputProc;
            SharpDX.RawInput.Device.RegisterDevice(SharpDX.Multimedia.UsagePage.Generic, SharpDX.Multimedia.UsageId.GenericKeyboard, DeviceFlags.Remove);
        }

        internal static void PollState()
        {
            Controller game_pad = new Controller(UserIndex.One);

            Gamepad gps;
            if (!game_pad.IsConnected) {
                gps = new Gamepad();
                gps.LeftThumbX = 0;
                if (global_key_state[key_button_map[42]]) gps.LeftThumbX += 0x4000;
                if (global_key_state[key_button_map[41]]) gps.LeftThumbX -= 0x4000;
                gps.LeftThumbY = 0;
                if (global_key_state[key_button_map[43]]) gps.LeftThumbY += 0x4000;
                if (global_key_state[key_button_map[44]]) gps.LeftThumbY -= 0x4000;
                gps.Buttons = 0;
                if (global_key_state[key_button_map[1]]) gps.Buttons |= GamepadButtonFlags.A;
                if (global_key_state[key_button_map[2]]) gps.Buttons |= GamepadButtonFlags.B;
                if (global_key_state[key_button_map[3]]) gps.Buttons |= GamepadButtonFlags.X;
                if (global_key_state[key_button_map[4]]) gps.Buttons |= GamepadButtonFlags.Y;
            } else {
                gps = game_pad.GetState().Gamepad;
            }

            for (int i = state_buffer.Length - 2; i >= 0; i--) {
                state_buffer[i + 1] = state_buffer[i];
            }

            int v = 0;
            for (int i = 0; i < button_settings.Length; i++) {
                if (IsButtonDown(gps, button_settings[i].Map)) v |= 0x1 << i;
            }
            state_buffer[0] = v;

            for (int i = 0; i < button_settings.Length; i++) {
                ButtonSetting bs = button_settings[i];
                int interval = 0;
                if (bs.IntervalRef == 2) {
                    interval = bs.Interval2;
                } else if (bs.IntervalRef == 1) {
                    interval = bs.Interval1;
                }
                if (interval > 0) {
                    int state_mask = 0x1 << i;
                    if ((state_buffer[0] & state_mask) != 0) {
                        bool on_interval = true;
                        for (int j = 1; j < interval; j++) {
                            if ((state_buffer[j] & state_mask) == 0) {
                                on_interval = false;
                                break;
                            }
                        }
                        if (on_interval) {
                            state_buffer[1] &= ~state_mask;
                            bs.IntervalRef = 2;
                        }
                    } else {
                        bs.IntervalRef = 1;
                    }
                    button_settings[i] = bs;
                }
            }
        }

        public static void MapButton(int button, int index)
        {
            button_settings[index].Map = (byte)button;
        }

        public static void MapKeyToButton(int key, int button)
        {
            key_button_map[button] = (byte)key;
        }

        public static void SetInputInterval(int index, int time1, int time2)
        {
            ButtonSetting bs = button_settings[index];
            bs.Interval1 = (byte)time1;
            bs.Interval2 = (byte)time2;
            if (time1 != 0 && time2 != 0) {
                bs.IntervalRef = 1;
            } else {
                bs.IntervalRef = 0;
            }
            button_settings[index] = bs;
        }

        public static int GetState()
        {
            return state_buffer[0];
        }

        public static int GetPreviousState()
        {
            return state_buffer[1];
        }
    }
}
