﻿using System;
using System.Collections.Generic;
using System.Drawing;
using GustFront;
using SharpDX.Direct3D9;
using SharpDX.XACT3;

namespace RVII
{
    public abstract class ContentInfo<T> : Element
    {
        public T Instance { get; set; }

        protected ContentInfo(string tagName)
            : base(tagName)
        {
        }

        public override Element Duplicate(string newTag)
        {
            throw new NotImplementedException();
        }
    }

    public sealed class TextureInfo : ContentInfo<Texture>
    {
        public TextureInfo(string tagName)
            : base(tagName)
        {
        }
    }

    public class ShaderInfo : ContentInfo<Effect>
    {
        public ShaderInfo(string tagName)
            : base(tagName)
        {
        }
    }

    public class GraphicsObjects : IGraphicsDevice
    {
        private Sprite mySpriteBatch = null;

        public GraphicsObjects(Sprite sprite_batch)
        {
            mySpriteBatch = sprite_batch;
        }

        public Sprite SpriteBatch { get { return mySpriteBatch; } }
    }

    public interface IRenderable
    {
        TextureInfo TextureInfo { get; }
    }

    public struct PointWrapper : IHasMember
    {
        int X;
        int Y;

        public static PointWrapper Zero { get { return new PointWrapper(); } }

        public Point ToPoint() { return new Point(X, Y); }

        public PointWrapper(int x, int y)
        {
            this.X = x;
            this.Y = y;
        }

        public PointWrapper(Point location)
        {
            this.X = location.X;
            this.Y = location.Y;
        }

        public override bool Equals(object obj)
        {
            return ToPoint().Equals(obj);
        }

        public bool Equals(PointWrapper other)
        {
            return ToPoint().Equals(other.ToPoint());
        }

        public static bool operator ==(PointWrapper a, PointWrapper b)
        {
            return a.ToPoint() == b.ToPoint();
        }

        public static bool operator !=(PointWrapper a, PointWrapper b)
        {
            return a.ToPoint() != b.ToPoint();
        }

        public override int GetHashCode()
        {
            return ToPoint().GetHashCode();
        }

        public override string ToString()
        {
            return ToPoint().ToString();
        }

        public bool GetMember(string name, ScriptValue[] @params, out ScriptValue result)
        {
            if (Util.CaseInsensitiveEquals(name, "X")) {
                result = X;
            } else if (Util.CaseInsensitiveEquals(name, "Y")) {
                result = Y;
            } else {
                result = null;
                return false;
            }
            return true;
        }

        public bool SetMember(string name, ScriptValue[] @params)
        {
            if (Util.CaseInsensitiveEquals(name, "X")) {
                X = @params[0].AsInt32();
            } else if (Util.CaseInsensitiveEquals(name, "Y")) {
                Y = @params[1].AsInt32();
            } else {
                return false;
            }
            return true;
        }
    }

    public sealed class WaveInfo : ContentInfo<WaveBank>
    {
        public WaveInfo(string tagName)
            : base(tagName)
        {
        }
    }

    public class CueWrapper : IHasMember
    {
        private Cue c = null;

        public CueWrapper(Cue c) { this.c = c; }

        public bool GetMember(string name, ScriptValue[] @params, out ScriptValue result)
        {
            result = null;

            switch (name) {
                case "IsCreated":
                    result = c.State == CueState.CreateD;
                    return true;
                case "IsPaused":
                    result = c.State == CueState.Paused;
                    return true;
                case "IsPlaying":
                    result = c.State == CueState.Playing;
                    return true;
                case "IsPrepared":
                    result = c.State == CueState.Prepared;
                    return true;
                case "IsPreparing":
                    result = c.State == CueState.Preparing;
                    return true;
                case "IsStopped":
                    result = c.State == CueState.Stopped;
                    return true;
                case "IsStopping":
                    result = c.State == CueState.Stopping;
                    return true;
                case "Name":
                    result = c.Properties.CueProperties.FriendlyName;
                    return true;
                case "GetVariable":
                    result = c.GetVariable(@params[0].AsInt16());
                    //result = c.GetVariable(@params[0].AsString());
                    return true;
                case "SetVariable":
                    c.SetVariable(@params[0].AsInt16(), @params[1].AsSingle());
                    //c.SetVariable(@params[0].AsString(), @params[1].AsSingle());
                    return true;
                case "Play":
                    c.Play();
                    return true;
                case "Pause":
                    c.Pause(true);
                    return true;
                case "Resume":
                    c.Pause(false);
                    return true;
                case "Stop":
                    c.Stop((StopFlags)@params[0].AsInt32());
                    return true;
            }

            return false;
        }

        public bool SetMember(string name, ScriptValue[] @params)
        {
            return false;
        }
    }

    public class FontInfo : ContentInfo<System.Drawing.Font>
    {
        private static FontInfo myDefaultFont = null;
        public static FontInfo DefaultFont { get { return myDefaultFont; } }

        static FontInfo()
        {
            myDefaultFont = new FontInfo("F_DEF");
            myDefaultFont.Instance = new System.Drawing.Font(
                System.Drawing.FontFamily.GenericMonospace, 13.0F);
        }

        private Element mySource = null;

        public FontInfo(string tagName)
            : base(tagName)
        {
        }

        public Element Source
        {
            get { return mySource; }
            internal set { mySource = value; }
        }
    }
}
